/*
Copyright (C) 2011 by Tomasz Konojacki

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include "microutf8.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

int32_t utf8_strlen(uint8_t *str)
{
	/* This function doesn't properly check validity of UTF-8 character 
	sequence, it is supposed to use only with valid UTF-8 strings. */
	
	int32_t character_count = 0;
	int32_t i = 0; /* Counter used to iterate over string. */
	uint8_t maybe_bom[3];
	
	/* If there is UTF-8 BOM ignore it. */
	if (strlen(str) > 2)
	{
		strncpy(maybe_bom, str, 3);
		if (strcmp(maybe_bom, UTF8_BOM) == 0)
			i += 3;
	}
	
	while(str[i])
	{
		/* If bit pattern begins with 0 we have ascii character. */ 
		if (str[i] >> 7 == 0)
			++character_count;
		/* If bit pattern begins with 11 it is beginning of UTF-8 byte
		sequence. */
		else if (str[i] >> 6 == 3)
			++character_count;
		/* If bit pattern begins with 10 it is middle of utf-8 byte sequence. */
		else if (str[i] >> 6 == 2)
			;
		/* In any other case this is not valid UTF-8. */
		else
			return UTF8_INVALID_SEQUENCE;
		++i;
	}
	return character_count;	
}

int8_t utf8_get_nth_char(uint8_t *str, uint8_t *target, uint32_t n, uint8_t max_length)
{
	uint8_t character_length = 0;
	int32_t character_count = 0;
	int32_t i = 0; /* Counter used to iterate over string. */
	int32_t _i = 0; /* Second counter, used to calculate size of UTF-8 character
	and to iterate over target character array */
	uint8_t maybe_bom[3];
	
	if (max_length == 0)
		return UTF8_NOT_ENOUGH_SPACE;
		
	/* If there is UTF-8 BOM ignore it. */
	if (strlen(str) > 2)
	{
		strncpy(maybe_bom, str, 3);
		if (strcmp(maybe_bom, UTF8_BOM) == 0)
			i += 3;
	}
	
	while(str[i])
	{
		/* If bit pattern begins with 0 we have ascii character. */ 
		if (str[i] >> 7 == 0)
		{
			if (++character_count == n)
			{
				/* That's the character that we were after. */
				if (max_length < 2)
					return UTF8_NOT_ENOUGH_SPACE;
				character_length = 1;
				target[0] = str[i];
				target[1] = 0;
				return 1;
			}
		}
		/* If bit pattern begins with 11 it is beginning of UTF-8 byte
		sequence. */
		else if (str[i] >> 6 == 3)
		{
			if (++character_count == n)
			{
				/* That's the character that we were after. */
				character_length = 1;
				_i = i;
				/* Check size (in bytes) of character. */
				while (str[++_i] >> 6 == 2)
					++character_length;
				/* Check if there is enough space for character. */
				if (character_length + 1 > max_length)
					return UTF8_NOT_ENOUGH_SPACE;
				/* This sequence must be longer than one byte. */
				if (character_length == 1)
					return UTF8_INVALID_SEQUENCE;
				/* Write character to target array. */
				for (_i = 0; _i < character_length; ++_i)
					target[_i] = str[i++];
				/* End string which contains character with 0. */
				target[_i++] = 0;
				return 1;		
			}
		}
		/* If bit pattern begins with 10 it is middle of UTF-8 byte sequence. */
		else if (str[i] >> 6 == 2)
			;
		/* In any other case this is not valid UTF-8. */
		else
			return UTF8_INVALID_SEQUENCE;
		++i;
	}
	/* It should not happen. */
	return 0;
}

